library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;


entity fp_div is
	generic
	(
		exp_bits	:	integer	:=	EXP_WIDTH;
		man_bits	:	integer	:=	MAN_WIDTH
	);
	port
	(
		op1		:	in	std_logic_vector(exp_bits+man_bits+1 downto 0);
		op2		:	in	std_logic_vector(exp_bits+man_bits+1 downto 0);
		READY		:	in	std_logic;
		EXCEPTION_IN	:	in	std_logic;
		CLK		:	in	std_logic;
		RESULT		:	out	std_logic_vector(exp_bits+man_bits+2 downto 0);
		EXCEPTION_OUT	:	out	std_logic																					:=	'0';
		DONE		:	out	std_logic																					:=	'0'
	);
end entity;

architecture fp_div_arch of fp_div is

component divider is
	generic
	(
		bits	:	integer
	);
	port
	(
		--inputs
		X	:	in	std_logic_vector(2*bits-1 downto 0);
		Y	:	in	std_logic_vector(2*bits-1 downto 0);
		CLK	:	in	std_logic;
		--outputs
		RESULT	:	out	std_logic_vector(2*bits+1 downto 0)
	);	
end component;
	
component parameterized_or_gate is
	generic
	(
		bits	:	integer
	);
	port
	(
		A	:	in	std_logic_vector(bits-1 downto 0);
		O	:	out	std_logic
	);
end component;

component parameterized_and_gate is
	generic
	(
		bits	:	integer
	);
	port
	(
		A	:	in	std_logic_vector(bits-1 downto 0);
		O	:	out	std_logic
	);
end component;

type signArray is array(PIPELINE_FP downto 0) of std_logic;
type expArray is array(PIPELINE_FP downto 0) of std_logic_vector(exp_bits downto 0);
type rdyArray is array(PIPELINE_FP downto 0) of std_logic;
type excArray is array(PIPELINE_FP downto 0) of std_logic;
type corArray is array(PIPELINE_FP downto 0) of std_logic;

constant  bias_val :	integer	:=	(2**(exp_bits-1))-1;
constant  m : integer := (man_bits+1+1)/2;  

signal	s1 : std_logic;
signal	s2 : std_logic;
signal  signs : signArray;
signal	sign : std_logic;
signal	sign0 : std_logic;
	
signal	e1 : std_logic_vector(exp_bits-1 downto 0);
signal	e2 : std_logic_vector(exp_bits-1 downto 0);
signal	e1_ext : std_logic_vector(exp_bits downto 0);
signal	e2_ext : std_logic_vector(exp_bits downto 0);
signal	exps_unbiased:	expArray;
signal  exp_unbiased : std_logic_vector(exp_bits downto 0);
signal  exp_zero : std_logic_vector(exp_bits-1 downto 0);
signal  sel : std_logic;
signal	exp_unbiased_shift:	std_logic_vector(exp_bits downto 0);
signal	exp_unbiased_shift1:	std_logic_vector(exp_bits downto 0);
signal	exp_unbiased_shift1_ext:	std_logic_vector(exp_bits+1 downto 0);
signal	bias:	std_logic_vector(exp_bits-1 downto 0);
signal	exp :	std_logic_vector(exp_bits-1 downto 0);
	
signal	f1 : std_logic_vector(man_bits downto 0);
signal	f2 :std_logic_vector(man_bits downto 0);
signal  f1_zero_ext : std_logic_vector(2*m-man_bits-1 downto 0);
signal  f2_zero_ext : std_logic_vector(2*m-man_bits-1 downto 0);
signal	f1_ext1 : std_logic_vector(2*m downto 0);
signal	f2_ext1 : std_logic_vector(2*m downto 0);
signal	f1_ext : std_logic_vector(2*m-1 downto 0);
signal	f2_ext : std_logic_vector(2*m-1 downto 0);
signal	man_l : std_logic_vector(2*m+1 downto 0);
signal	man_s : std_logic_vector(man_bits+2 downto 0);
signal	man_unshift : std_logic_vector(man_bits+1 downto 0);
signal	man_shift : std_logic_vector(man_bits+1 downto 0);
signal	man : std_logic_vector(man_bits+1 downto 0);
signal	man_1 : std_logic_vector(man_bits+1 downto 0);
signal	man_normal : std_logic_vector(man_bits+1 downto 0);
signal	man_normal1 : std_logic_vector(man_bits downto 0);
    
signal	value : std_logic_vector(exp_bits+man_bits+2 downto 0);
signal	res_zero : std_logic_vector(exp_bits+man_bits+2 downto 0);
signal	res : std_logic_vector(exp_bits+man_bits+2 downto 0);
	
signal  dividend_exp_all_zero : std_logic;
signal  dividend_exp_zero : std_logic;
signal  dividend_exp_max : std_logic;
signal  divisor_exp_all_zero : std_logic;
signal  divisor_exp_zero : std_logic;
signal  divisor_exp_max : std_logic;
signal  corners : corArray;
signal  corner : std_logic;
signal  corner0 : std_logic;
signal  exp_unbiased_shift1_or_zero : std_logic;
signal  overflow : std_logic;
signal  underflow : std_logic;
signal  special : std_logic;

signal  excs : excArray;
signal  exc : std_logic;
signal  exc0 : std_logic;

signal  rdys : rdyArray;
signal  rdy : std_logic;
signal  rdy0 : std_logic;
   	
	
begin

mantissa_divider : divider
generic map
(
	bits => m
)
port map
(
        X => f1_ext,
        Y => f2_ext,
        CLK => clk,
        RESULT => man_l
);    

all_zero_underflow: parameterized_or_gate
	generic map
	(
		bits	=>	exp_bits-1
	)
	port map
	(
		A	=>	exp_unbiased_shift1_ext(exp_bits-1 downto 1),
		O	=>	exp_unbiased_shift1_or_zero
	);

all_zero_exp_dividend: parameterized_or_gate
	generic map
	(
		bits	=>	exp_bits
	)
	port map
	(
		A	=>	e1,
		O	=>	dividend_exp_all_zero
	);

all_zero_exp_divisor: parameterized_or_gate
	generic map
	(
		bits	=>	exp_bits
	)
	port map
	(
		A	=>	e2,
		O	=>	divisor_exp_all_zero
	);

all_one_exp_dividend: parameterized_and_gate
	generic map
	(
		bits	=>	exp_bits
	)
	port map
	(
		A	=>	e1,
		O	=>	dividend_exp_max
	);

all_one_exp_divisor: parameterized_and_gate
	generic map
	(
		bits	=>	exp_bits
	)
	port map
	(
		A	=>	e2,
		O	=>	divisor_exp_max
	);


----------------------------------------------------------
--              Asynchronous Assignments                --
----------------------------------------------------------
	s1 <= op1(exp_bits+man_bits+1);
	s2 <= op2(exp_bits+man_bits+1);
	e1 <= op1(exp_bits+man_bits downto man_bits+1);
	e2 <= op2(exp_bits+man_bits downto man_bits+1);
	f1 <= op1(man_bits downto 0);
	f2 <= op2(man_bits downto 0);


	-- get the exponent
	-- compute the exponent difference e1-e2
	e1_ext <= '0' & e1;
	e2_ext <= '0' & e2;
	exps_unbiased(0) <= e1_ext - e2_ext;
    
	-- adjust exponent, either non-change or decrement by one
	-- for mantissa result as 1.xxx, exponent result doesn't change
	-- for mantissa result as 0.1xx, exponent result decrement
	exp_zero <= (others=>'0');
	exp_unbiased_shift <= exp_unbiased - (exp_zero & not(sel));
	exp_unbiased_shift1_ext <= exp_unbiased_shift1(exp_bits downto 1) & '0' & exp_unbiased_shift1(0);
	-- add bias to the exponent different after adjust
	bias <= conv_std_logic_vector(bias_val, exp_bits);
	exp <= exp_unbiased_shift1(exp_bits-1 downto 0) + bias;

    
	-- get the mantissa
	-- zero padding f1, f2 and make it even bits for "divider" to use
	f1_zero_ext <= (others=>'0');
	f1_ext1 <= f1 & f1_zero_ext;
	f1_ext <= f1_ext1(2*m downto 1);
	f2_zero_ext <= (others=>'0');
	f2_ext1 <= f2 & f2_zero_ext;
	f2_ext <= f2_ext1(2*m downto 1);

	-- get the correct mantissa from the result of "divider"
	-- for result as 1.xxx, unshift
	-- for result as 0.1xx, mantissa left  by one bit, exponent decrement
	-- man is the correct mantissa with the implied '1'
	man_s <= man_l(2*m+1 downto 2*m-man_bits-1);
	man_unshift <= man_s(man_bits+2 downto 1);
	man_shift <= man_s(man_bits+1 downto 0);
	man <= man_shift when man_l(2*m+1)='0' else
	       man_unshift;
	man_normal <= man_1(man_bits+1 downto 0);
			    
	
	-- special cases
	-- detect divide by zero, inf, nan, denormal either dividend and divisor
	dividend_exp_zero <= not(dividend_exp_all_zero);
	divisor_exp_zero <= not(divisor_exp_all_zero);
	corners(0) <= dividend_exp_max or divisor_exp_max or divisor_exp_zero or dividend_exp_zero;
    
	-- overflow : 01xx..xx
	overflow <= not( exp_unbiased_shift1(exp_bits) ) and exp_unbiased_shift1(exp_bits-1);
	-- underflow : 10xx..xx or
	underflow <= ( exp_unbiased_shift1(exp_bits) and not( exp_unbiased_shift1(exp_bits-1) ) ) or ( exp_unbiased_shift1(exp_bits) and exp_unbiased_shift1(exp_bits-1) and not(exp_unbiased_shift1_or_zero) );
    
	special <= overflow or underflow or exc0 or corner0;


	-- result
    	value(exp_bits+man_bits+2) <= sign0;
	value(exp_bits+man_bits+1 downto man_bits+2) <=	exp;
	value(man_bits+1 downto 0) <= '1' & man_normal1;
	-- select between special and non-special
	res_zero <= (others=>'0');
    	res <= res_zero when special='1' else
	       value;


	signs(0) <= s1 xor s2;
	-- ready and done signal
	rdys(0) <= ready;
	-- exc signal
	excs(0) <= exception_in;

    
	sign_regs : for i in 1 to PIPELINE_FP generate
		signs(i) <= signs(i-1) when rising_edge(clk);
	end generate;
	           
	exp_regs : for i in 1 to PIPELINE_FP generate
		exps_unbiased(i) <= exps_unbiased(i-1) when rising_edge(clk);
	end generate;
    
	rdy_regs : for i in 1 to PIPELINE_FP generate
		rdys(i) <= rdys(i-1) when rising_edge(clk);
	end generate;

	exc_regs : for i in 1 to PIPELINE_FP generate
		excs(i) <= excs(i-1) when rising_edge(clk);
	end generate;

	cor_regs : for i in 1 to PIPELINE_FP generate
		corners(i) <= corners(i-1) when rising_edge(clk);
	end generate;


	done <= rdy0;
	exception_out <= special;
	result <= res;


----------------------------------------------------------
--               Synchronous Assignments                --
----------------------------------------------------------
synchro: process (CLK)
begin
	if(rising_edge(CLK)) then
        	sign <= signs(PIPELINE_FP);
        	sign0 <= sign;
            
	        sel <= man_l(2*m+1);
	        exp_unbiased <= exps_unbiased(PIPELINE_FP);
	        exp_unbiased_shift1 <= exp_unbiased_shift;
            
	        man_1 <= man;
	        man_normal1 <= man_normal(man_bits downto 0);

	        rdy <= rdys(PIPELINE_FP);
	        rdy0 <= rdy;

	        exc <= excs(PIPELINE_FP);
	        exc0 <= exc;
            
	        corner <= corners(PIPELINE_FP);
	        corner0 <= corner;
	end if;
end process;

end fp_div_arch; 