library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;


entity fp_sqrt is
	generic
	(
		exp_bits	:	integer	:=	EXP_WIDTH;
		man_bits	:	integer	:=	MAN_WIDTH
	);
	port
	(
		--inputs
		OP	:		in	std_logic_vector(exp_bits+man_bits+1 downto 0);
		READY	:		in	std_logic;
		EXCEPTION_IN	:	in	std_logic;
		CLK	:		in	std_logic;
		--outputs
		RESULT	:		out	std_logic_vector(exp_bits+((man_bits+3)/4+1)*4+1 downto 0);
		EXCEPTION_OUT	:	out	std_logic																					:=	'0';
		DONE	:		out	std_logic																					:=	'0'
	);
end fp_sqrt;


architecture fp_sqrt_arch of fp_sqrt is

component sq is
	generic
	(
		bits		:	integer
	);
	port
	(
		--inputs
        	odd_in    	:	in      std_logic;
		Y_in		:	in	std_logic_vector(4*bits-1 downto 0); 
		CLK		:	in	std_logic;
		--outputs
		res		:	out	std_logic_vector(4*bits-1 downto 0) 
	);	
end component;
	
component parameterized_or_gate is
	generic
	(
		bits	:	integer
	);
	port
	(
		A	:	in	std_logic_vector(bits-1 downto 0);
		O	:	out	std_logic
	);
	end component;

component parameterized_and_gate is
	generic
	(
		bits	:	integer
	);
	port
	(
		A	:	in	std_logic_vector(bits-1 downto 0);
		O	:	out	std_logic
	);
end component;



--======================================================--
--                      SIGNALS                         --
--======================================================--
constant    sq_bits	:integer	:= (man_bits+3)/4+1;
constant    f_bits	: integer	:= sq_bits * 4;

type rdyArray  is array(PIPELINE_FP downto 0) of std_logic;	
type excArray  is array(PIPELINE_FP downto 0) of std_logic;	
type signArray is array(PIPELINE_FP downto 0) of std_logic;	
type expArray  is array(PIPELINE_FP downto 0) of std_logic_vector(exp_bits-1 downto 0);	
	
signal	s	:	std_logic;
signal	e	:	std_logic_vector(exp_bits-1 downto 0);
signal	f	:	std_logic_vector(man_bits-1 downto 0);

signal	bias:	std_logic_vector(exp_bits-1 downto 0);
signal	e_unbiased:	std_logic_vector(exp_bits-1 downto 0);
signal	exp_unbiased:	std_logic_vector(exp_bits-1 downto 0);
signal	exp:	std_logic_vector(exp_bits-1 downto 0);

signal  man_in  :   std_logic_vector(f_bits-1 downto 0);
signal  man_out  :   std_logic_vector(f_bits-1 downto 0);
	
signal	exp_zero	:	std_logic_vector(exp_bits-1 downto 0);
signal	man_zero	:	std_logic_vector(f_bits downto 0);
signal	e_zero	:	std_logic;
signal  e_max   :   std_logic;
signal  corner :   std_logic;
    
signal  rdys : rdyArray;
signal  signs : signArray;
signal  exps : expArray;
	
signal	exc1	:	std_logic;
signal	exc2	:	std_logic;
signal  excs : excArray;

signal	e_odd	:	std_logic;

	
begin
----------------------------------------------------------
--               Component Instantiation                --
----------------------------------------------------------
		
all_zero_exp: parameterized_or_gate
	generic map
	(
		bits	=>	exp_bits
	)
	port map
	(
		A	=>	e,
		O	=>	e_zero
	);
		
all_one_exp: parameterized_and_gate
	generic map
	(
		bits	=>	exp_bits
	)
	port map
	(
		A	=>	e,
		O	=>	e_max
	);		
		
man_comp :	 sq 
	generic map
	(
		bits	=>	sq_bits
	)
	port map
	(
		--inputs
		odd_in	=>	e_odd,
		Y_in	=>	man_in,
		CLK	=>	clk,
		res	=>	man_out 
	);	
		

		
----------------------------------------------------------
--              Asynchronous Assignments                --
----------------------------------------------------------
bias(exp_bits-1)<=	'0';
bias(exp_bits-2 downto 0)	<=	(others=>'1');
	
s	<=	OP(exp_bits+man_bits+1);
e	<=	OP(exp_bits+man_bits downto man_bits+1);
f	<=	OP(man_bits-1 downto 0);
	
man_in(f_bits-1 downto f_bits-man_bits) <= f(man_bits-1 downto 0);
man_in(f_bits-man_bits-1 downto 0) <= (others=>'0');

e_unbiased <= e - bias;
e_odd <= e_unbiased(0);
exp_unbiased <= e_unbiased(exp_bits-1) & e_unbiased(exp_bits-1 downto 1);
exp <= exp_unbiased + bias;

exp_zero <= (others=>'0');
man_zero <= (others=>'0');
		
result(exp_bits+f_bits+1) <= signs(PIPELINE_FP) when excs(PIPELINE_FP)='0' else '0';
result(exp_bits+f_bits downto f_bits+1) <= exps(PIPELINE_FP) when excs(PIPELINE_FP)='0' else exp_zero;
result(f_bits downto 0) <= '1' & man_out when excs(PIPELINE_FP)='0' else	man_zero;

DONE		<=	rdys(PIPELINE_FP);
EXCEPTION_OUT <= excs(PIPELINE_FP);			

rdys(0) <= READY;
rdy_regs : for i in 1 to PIPELINE_FP generate
	rdys(i) <= rdys(i-1) when rising_edge(clk);
end generate;

signs(0) <= s;
sign_regs : for i in 1 to PIPELINE_FP generate
        signs(i) <= signs(i-1) when rising_edge(clk);
end generate;

exps(0) <= exp;
exp_regs : for i in 1 to PIPELINE_FP generate
        exps(i) <= exps(i-1) when rising_edge(clk);
end generate;
    
excs(2) <= exc2;
exc_regs : for i in 3 to PIPELINE_FP generate
	excs(i) <= excs(i-1) when rising_edge(clk);
end generate;

	
----------------------------------------------------------
--               Synchronous Assignments                --
----------------------------------------------------------
synchro: process (CLK)
begin
	if(rising_edge(CLK)) then
	        corner <= NOT(e_zero) or e_max;
	        exc1 <=	EXCEPTION_IN or s;
	        exc2 <=	corner OR exc1;
	end if;
end process;

end fp_sqrt_arch; 